@@ -6,6 +6,7 @@ import sys
 import base58
 import base64
 from urllib.parse import quote
+from pathlib import Path
 
 import asyncio
 import aiofiles
@@ -185,19 +186,25 @@ class SubAPI(object):
         if isinstance(headers, dict):
             kwargs['headers'] = headers
 
-        async with getattr(session, method)(url, **kwargs) as response:
-            async for raw_message in response.content:
-                message = decode_json(raw_message)
+        try:
+            async with getattr(session, method)(url, **kwargs) as response:
+                async for raw_message in response.content:
+                    message = decode_json(raw_message)
 
-                if message is not None:
-                    if 'Message' in message and 'Code' in message:
-                        raise aioipfs.APIError(code=message['Code'],
-                                               message=message['Message'],
-                                               http_status=response.status)
-                    else:
-                        yield message
+                    if message is not None:
+                        if 'Message' in message and 'Code' in message:
+                            raise aioipfs.APIError(code=message['Code'],
+                                                   message=message['Message'],
+                                                   http_status=response.status)
+                        else:
+                            yield message
 
-                await asyncio.sleep(0)
+                    await asyncio.sleep(0)
+        except asyncio.CancelledError as err:
+            if new_session is True:
+                await session.close()
+
+            raise err
 
         if new_session is True:
             await session.close()
@@ -551,7 +558,7 @@ class ConfigAPI(SubAPI):
 
     async def show(self):
         """ Outputs IPFS config file contents """
-        return await self.fetch_text(self.url('config/show'))
+        return await self.fetch_json(self.url('config/show'))
 
     async def replace(self, configpath):
         """ Replaces the IPFS configuration with new config file
@@ -756,6 +763,12 @@ class DiagAPI(SubAPI):
         return await self.fetch_text(self.url('diag/cmds/clear'))
 
 
+class FileAPI(SubAPI):
+    async def ls(self, path):
+        params = {ARG_PARAM: path}
+        return await self.fetch_json(self.url('file/ls'), params=params)
+
+
 class FilesAPI(SubAPI):
     async def cp(self, source, dest):
         params = quote_args(source, dest)
@@ -764,7 +777,7 @@ class FilesAPI(SubAPI):
 
     async def chcid(self, path, cidversion):
         params = {ARG_PARAM: path, 'cid-version': str(cidversion)}
-        return await self.fetch_text(self.url('files/cp'), params=params)
+        return await self.fetch_text(self.url('files/chcid'), params=params)
 
     async def flush(self, path):
         params = {ARG_PARAM: path}
@@ -1098,6 +1111,18 @@ class ObjectAPI(SubAPI):
         params = {ARG_PARAM: objkey}
         return await self.fetch_json(self.url('object/stat'), params=params)
 
+    async def diff(self, obja, objb, verbose=False):
+        """
+        Display the diff between two ipfs objects.
+        """
+
+        params = quote_dict({
+            ARG_PARAM: [obja, objb],
+            'verbose': boolarg(verbose)
+        })
+
+        return await self.fetch_json(self.url('object/diff'), params=params)
+
     async def get(self, objkey):
         params = {ARG_PARAM: objkey}
         return await self.fetch_json(self.url('object/get'), params=params)
@@ -1233,7 +1258,7 @@ class PubSubAPI(SubAPI):
         return await self.post(self.url('pubsub/pub'),
                                None, params=quote_args(topic, data))
 
-    async def sub(self, topic, discover=True):
+    async def sub(self, topic, discover=True, timeout=None, read_timeout=None):
         """
         Subscribe to messages on a given topic.
 
@@ -1253,8 +1278,8 @@ class PubSubAPI(SubAPI):
                 method='post',
                 headers={'Connection': 'Close'},
                 new_session=True,
-                timeout=60.0 * 60 * 24 * 8,
-                read_timeout=0,
+                timeout=timeout if timeout else 60.0 * 60 * 24 * 8,
+                read_timeout=read_timeout if read_timeout else 0,
                 params=params):
             try:
                 converted = self.decode_message(message)
@@ -1293,8 +1318,10 @@ class RepoAPI(SubAPI):
             'quiet': boolarg(quiet),
             'stream-errors': boolarg(streamerrors)
         }
-        return await self.fetch_text(self.url('repo/gc'),
-                                     params=params)
+
+        async for ent in self.mjson_decode(self.url('repo/gc'),
+                                           params=params):
+            yield ent
 
     async def verify(self):
         return await self.fetch_json(self.url('repo/verify'))
@@ -1403,10 +1430,8 @@ class CoreAPI(SubAPI):
         params = {}
         switchlist = [
             'fscache',
-            'hidden',
             'nocopy',
             'only_hash',
-            'offline',
             'pin',
             'quiet',
             'quieter',
@@ -1527,14 +1552,19 @@ class CoreAPI(SubAPI):
         :param int cid_version: CID version
         """
 
+        hidden = kwargs.pop('hidden', False)
+        ignrulespath = kwargs.pop('ignore_rules_path', None)
         params = self._build_add_params(**kwargs)
 
+        descriptors = []
         all_files = []
         for fitem in files:
             if isinstance(fitem, list):
                 all_files += fitem
             elif isinstance(fitem, str):
                 all_files.append(fitem)
+            elif isinstance(fitem, Path):
+                all_files.append(str(fitem))
 
         # Build the multipart form and add the files/directories
         with multi.FormDataWriter() as mpwriter:
@@ -1548,7 +1578,9 @@ class CoreAPI(SubAPI):
                     continue
 
                 if os.path.isdir(filepath):
-                    dir_listing = multi.DirectoryListing(filepath)
+                    dir_listing = multi.DirectoryListing(
+                        filepath, hidden=hidden,
+                        ignrulespath=ignrulespath)
                     names = dir_listing.genNames()
                     for entry in names:
                         await asyncio.sleep(0)
@@ -1572,11 +1604,14 @@ class CoreAPI(SubAPI):
                             pay.set_content_disposition(
                                 'form-data', name='file', filename=_name)
                             mpwriter.append_payload(pay)
+
+                        descriptors.append(_fd)
                 else:
                     basename = os.path.basename(filepath)
+                    _fd = open(filepath, 'rb')
 
                     file_payload = payload.BytesIOPayload(
-                        open(filepath, 'rb'),
+                        _fd,
                         content_type='application/octet-stream',
                         headers={
                             'Abspath': filepath
@@ -1587,12 +1622,22 @@ class CoreAPI(SubAPI):
                                                          name='file',
                                                          filename=basename)
                     mpwriter.append_payload(file_payload)
+                    descriptors.append(_fd)
 
             if mpwriter.size == 0:
                 raise aioipfs.UnknownAPIError('Multipart is empty')
 
-            async for value in self.add_generic(mpwriter, params=params):
-                yield value
+            try:
+                async for value in self.add_generic(mpwriter, params=params):
+                    yield value
+            except Exception as err:
+                for fd in descriptors:
+                    try:
+                        fd.close()
+                    except Exception:
+                        continue
+
+                raise err
 
     async def commands(self):
         """ List all available commands."""
@@ -1698,6 +1743,94 @@ class CoreAPI(SubAPI):
         tar_future = loop.run_in_executor(None, extract)
         return await tar_future
 
+    async def getgen(self, objpath, dstdir='.', compress=False,
+                     compression_level=-1, archive=True, output=None,
+                     progress_callback=None, progress_callback_arg=None,
+                     sleept=0.05, chunk_size=16384):
+        """
+        Download IPFS objects.
+
+        :param str objpath: The base58 objpath of the object to retrieve
+        :param str dstdir: destination directory, current directory by default
+        :param bool compress: Compress the output with GZIP compression
+        :param str compression_level: The level of compression (1-9)
+        :param bool archive: Output a TAR archive
+        """
+
+        params = {
+            ARG_PARAM: objpath,
+            'compress': boolarg(compress),
+            'compression-level': str(compression_level),
+            'archive': boolarg(archive)
+        }
+
+        if isinstance(output, str):
+            params['output'] = output
+
+        archive_path = tempfile.mkstemp(prefix='aioipfs')[1]
+
+        # We read chunk by chunk the tar data coming from the
+        # daemon and use aiofiles to asynchronously write the data to
+        # the temporary file
+
+        read_so_far = 0
+        chunks = 0
+
+        try:
+            async with aiofiles.open(archive_path, 'wb') as fd:
+                async with self.driver.session.post(self.url('get'),
+                                                    params=params) as response:
+                    if response.status != 200:
+                        raise aioipfs.APIError(http_status=response.status)
+
+                    headers = response.headers
+                    clength = int(headers.get('X-Content-Length'))
+
+                    while True:
+                        chunk = await response.content.read(chunk_size)
+                        if not chunk:
+                            break
+
+                        chunks += 1
+                        read_so_far += len(chunk)
+
+                        yield 0, read_so_far, clength
+
+                        await fd.write(chunk)
+                        await asyncio.sleep(sleept)
+
+                    await response.release()
+        except GeneratorExit:
+            os.unlink(archive_path)
+            raise
+        except Exception:
+            os.unlink(archive_path)
+            raise
+        else:
+            def extract():
+                # Synchronous tar extraction runs in the executor
+                mode = 'r|gz' if compress is True else 'r|'
+                try:
+                    with tarfile.open(name=archive_path, mode=mode) as tf:
+                        tf.extractall(path=dstdir)
+                    os.unlink(archive_path)
+                    return True
+                except Exception as e:
+                    print(
+                        'Could not extract TAR file:', str(e), file=sys.stderr)
+                    os.unlink(archive_path)
+                    return False
+
+            # Run the tar extraction inside asyncio's threadpool
+            loop = asyncio.get_event_loop()
+            tar_future = loop.run_in_executor(None, extract)
+            res = await tar_future
+
+            if res is True:
+                yield 1, None, None
+            else:
+                yield -1, None, None
+
     async def ls(self, path, headers=False, resolve_type=True):
         """
         List directory contents for Unix filesystem objects.
@@ -1714,6 +1847,26 @@ class CoreAPI(SubAPI):
         }
         return await self.fetch_json(self.url('ls'), params=params)
 
+    async def ls_streamed(self, path, headers=False, resolve_type=True):
+        """
+        List directory contents for Unix filesystem objects (streamed)
+
+        :param str path: The path to the IPFS object(s) to list links from
+        :param bool headers: Print table headers (Hash, Size, Name)
+        :param bool resolve_type: Resolve linked objects to get their types
+        """
+
+        params = {
+            'resolve-type': boolarg(resolve_type),
+            'headers': boolarg(headers),
+            'stream': boolarg(True),
+            ARG_PARAM: path
+        }
+
+        async for value in self.mjson_decode(
+                self.url('ls'), params=params):
+            yield value
+
     async def mount(self, ipfspath, ipnspath):
         params = {
             'ipfs-path': ipfspath,
@@ -1752,6 +1905,12 @@ class CoreAPI(SubAPI):
 
         return await self.fetch_json(self.url('version'), params=params)
 
+    async def version_deps(self):
+        """
+        Shows information about dependencies used for build
+        """
+        return await self.fetch_json(self.url('version/deps'))
+
     async def dns(self, name, recursive=False):
         """
         Resolve DNS links
